// Copyright (c) 2010 Guillaume Lathoud
// MIT License

(function (/*global context object*/global) {

    function log() {
        var f = global.console && global.console.log;
        f && f.apply && f.apply( console, arguments);
    }
    function error() {
        var f = global.console && global.console.error;
        f && f.apply && f.apply( console, arguments);
    }

    // xxx
    log = function () {};

    global.tailopt = function (/* string | function*/ fname_or_fun, /* function | string */ fun_or_head, /*?string?*/ body) {

        var fname, arr, rx, s, head, new_head, new_body, o, ret;

        //

        if (typeof fname_or_fun === 'function') {
            return tailopt("", fname_or_fun);
        }
        fname = fname_or_fun;

        //

        if (typeof fun_or_head === 'function') {
            if (! fun_or_head.toString) {
                error('function.toString() not supported. Returning the function unchanged.');
                return fun_or_head;
            }

            s = fun_or_head.toString();
            rx = /^\s*(function\s*([\w]*?)\s*\(.*?\))\s*\{([\w\W]*?)\}\s*$/;
            if (! rx.test(s)) {
                global.rx = rx;
                global.s = s;
                log(s);
                error('unrecognized function source. Returning the function unchanged.');
                return fun_or_head;
            }
            arr = rx.exec(s);
            return tailopt(fname || arr[2], arr[1], arr[3]);
        }
        head = fun_or_head;

        //

        try {
            new_head = remove_comments(head);
            
            new_body = remove_comments(body);
            
            new_body = unroll_all_namedlets(new_head, new_body);
            
            if (fname) {
                o = try_unroll_top(new_head, new_body, fname);
                new_head = o.head;
                new_body = o.body;
            }

            log('xxx new head:', new_head);
            log('xxx new body:', new_body );
            
        } catch(e) {            

            log('tailopt() head:', head);
            log('tailopt() body.substr(0, 100):', body.substr(0, 100));
            error('tailopt() is returning the function unmodified because it caught an exception: ' + e);
            
            new_head = head;
            new_body = body;
        }

        return eval('(' + new_head + '{' + new_body + '})'); // xxx not necessary anymore? Ugly 'ret = (' because of IE
    };

    function remove_comments(s) {

        // xxx make this resistant to perverse cases with /* or // within strings or regexps
        // - concretely: we must leave strings and regexp intact
        // - a full JS parser may be needed
        //
        // However for most practical cases the current remove_comments() code should be sufficient!
        
        var a, b, ret = s;

        while (true) {
            a = ret.indexOf('/*');
            b = ret.indexOf('//');

            a = a < 0 ? +Infinity : a;
            b = b < 0 ? +Infinity : b;

            if (! ((a < +Infinity) || (b < +Infinity))) {
                break;
            }
            
            if (a < b) {  // Remove such a comment
                ret = ret.replace(/\/\*[\w\W]*?\*\//, '\n');
            } else { /* Remove such a comment */
                ret = ret.replace(/\/\/[\w\W]*?$/m, '');
            }
            
        }

        return ret;
    };

    function unroll_all_namedlets(head, body) {
        
        var new_body;

        while (true) {
            new_body = unroll_first_namedlet(head, body);
            if (new_body === body) {
                return body;
            }

            log('xxx unroll_all_namedlets new body:', new_body);

            body = new_body;
        }
    };

    function try_unroll_top(head, body, fname) {
        
        if (! (head && body && fname)) {
            return;
        }           

        var a, arr = get_vars_from_head(head), 
        tmp_arr, tmp_head, tmp_body,
        new_head, new_body;

        // Maybe the top function itself has a tail call?
        // If yes, we should unroll it. Let us just try.

        tmp_arr = new Array(arr.length);
        for (a = 0; a < arr.length; a++) {
            // xxx again, we should check for variable name collisions,
            // and even avoid it. For now we'll just use improbable variable names.
            tmp_arr[a] = '__0_' + arr[a] + '_0__';
        }
        tmp_head = "function(" + tmp_arr.join(',') + ')';
        tmp_body = "var " + fname + "; return (" + fname + " = function (" + arr.join(',') + "){" + body + "})(" + tmp_arr.join(",") + ");";

        new_head = tmp_head;
        new_body = unroll_first_namedlet(tmp_head, tmp_body);

        if (new_body.substring(new_body.indexOf('{') + 1, new_body.lastIndexOf('}')) !== body) {
            // Success -> return the modified implementation
            return { 
                head: new_head,
                body: new_body
            };
        } else {
            // Failure -> return the original implementation
            return {
                head: head,
                body: body
            };
        }
        
    };

    function unroll_first_namedlet(head, body) {

        var a, a_rx, arr, b, c, e, block,  blocks, mo, rx_prefix, rx_postfix, rx_tail, new_block, new_block_inner, rx, s, prefix, postfix, vars, new_body;

        // List variables

        vars = get_vars_from_head(head).concat(get_vars_from_body(body));

        log("xxx vars:", vars);
        
        // List blocks

        blocks = find_blocks(body);

        // Find the first namedlet

        // xxx in the future, allow () in the prefix and postfix, for now only parenthese-free expressions
        // xxx or: some time in the future, generate an error message for imbricated ()

        // We recognize two forms
        rx = [
            // First form
            {
                prefix:  /\breturn\s*function\s*(\w+)\s*(\([^\(\)]*?\))\s*$/,
                postfix: /^\s*(\([^\(\)]*?\))/
            },
            // First form, with spurious parenthese
            {
                prefix:  /\breturn\s*\(\s*function\s*(\w+)\s*(\([^\(\)]*?\))\s*$/,
                postfix: /^\s*\)\s*(\([^\(\)]*?\))/
            },
            // Second form, with an anonymous function expression
            { 
                prefix:  /\breturn\s*\(\s*(\w+)\s*=\s*function\s*(\([^\(\)]*?\))\s*$/,
                postfix: /^\s*\)\s*(\([^\(\)]*?\))/
            }
        ];
        
        for (a = 0; a < blocks.length; a++) {
            s = blocks[a][0];
            e = blocks[a][1];

            prefix = body.substring(0, s);
            block  = body.substring(s, e+1);
            postfix = body.substring(e+1);

            //         log('.');
            //         log('xxx prefix', prefix);
            //         log('xxx postfix', postfix);
            //         log('.');
            
            for ( a_rx = 0; a_rx < rx.length; a_rx++ ) {

                rx_prefix  = rx[a_rx].prefix;
                rx_postfix = rx[a_rx].postfix;

                if ( rx_prefix.test( prefix ) && rx_postfix.test( postfix ) ) {      
                    
                    var mo_prefix = rx_prefix.exec(prefix);
                    var fname = mo_prefix[1];
                    var fpar = get_vars_from_head(mo_prefix[2]);

                    var mo_postfix = rx_postfix.exec(postfix);
                    var fparval = get_vars_from_head(mo_postfix[1]);
                    
                    log("xxx named let block: ", body.substring(s, e + 1));
                    log("xxx fname:", fname);
                    log("xxx fpar:", fpar);
                    log("xxx fparval:", fparval);
                    
                    // xxx (maybe at the end) Make sure the variable names will not collide after unrolling
                    // --> if necessary, rename fpar AND variables in block body
                    // --> or at least detect collision and ask the user to rename his variables

                    var label = 'L_' + fname;


                    // Expand recursive conditionals, like:
                    //     return <expr_without_fname> ? <fname>(...) : <something_else>
                    // or:
                    //     return <expr_without_fname> ? <expr_without_fname> : <fname>(...)
                    // or:
                    //     return <expr_without_fname> ? <fname>(...)> : <fname>(...)

                    new_block_inner = expand_recursive_conditionals(block, fname);

                    // xxx some time in the future, generate an error message for imbricated ()
                    rx_tail   = new RegExp('\\breturn\\s+' + fname + '\\s*(\\([^\\(\\)]*\\))\\s*;');
                    log("xxx rx_tail.test:", rx_tail.test(new_block_inner));

                    while (rx_tail.test(new_block_inner)) {
                        mo = rx_tail.exec(new_block_inner);
                        var tailval = get_vars_from_head(mo[1]);
                        log('xxx tailval:', tailval);

                        var tailrepl = '{ ';
                        for (b = 0; b < tailval.length; b++) {
                            if (tailval[b] !== fpar[b]) {
                                tailrepl += fpar[b] + '_new = ' + tailval[b] + ';\n';
                            }
                        }
                        for (b = 0; b < tailval.length; b++) {
                            if (tailval[b] !== fpar[b]) {
                                tailrepl += fpar[b] + ' = ' + fpar[b] + '_new;\n';
                            }
                        }
                        tailrepl += 'continue ' + label + ';\n}';

                        new_block_inner = new_block_inner.substring(0, mo.index) + 
                            tailrepl + 
                            new_block_inner.substring(mo.index + mo[0].length);

                    }


                    arr = [];
                    for (b = 0; b < fpar.length; b++) {
                        
                        c = fpar[b];
                        if (fparval[b] !== undefined) {
                            c += ' = ' + fparval[b];
                        }
                        arr.push(c);

                        arr.push(fpar[b] + '_new');
                    }

                    new_block = 'var ' + arr.join(', ') + ';\n';
                    new_block += label + ': while (true) ' + new_block_inner; 

                    log('xxx new_block:', new_block);

                    // Put the new block in place

                    new_body = body.substring(0, s - mo_prefix[0].length) + new_block + body.substring(e + 1 + mo_postfix[0].length);


                    log('xxx new_body:', new_body);

                    return new_body;  // xxx fix trailing extra ;

                }
            }

        }

        // xxx error message: no tail block found!

        return body;
        
    };

    function get_vars_from_head(head) {
        
        var a, s, ret;

        s = /\((.*?)\)/.exec(head)[1].split(',');

        ret = [];
        for (a = 0; a < s.length; a++) {
            ret.push(/^\s*(.*?)\s*$/.exec(s[a])[1]);
        } 

        return ret;
    };

    function get_vars_from_body(body) {

        // xxx not done yet

        return [];
    };

    function find_blocks(s) {
        
        // Find all '{' and '}', and group them into blocks

        var last = s.length,
        open = -1, close = -1,
        close_arr = [],
        block_arr = [];
        
        while (true) {
            last--;
            if (open < 0) { open  = s.lastIndexOf('{', last); }
            if (close < 0) {  close = s.lastIndexOf('}', last); }
            
            if (open > close) {
                last = open;
                block_arr.push([open, close_arr.pop()]);
                open = -1;
                
            } else if (close > open) {
                last = close;
                close_arr.push(close);
                close = -1;
                
            } else {
                break;
            }
        }
        
        block_arr.reverse();
        return block_arr;
        
    };


    function expand_recursive_conditionals(block, fname) {
        var ret = block,
        rx = /\breturn\s+([^;\?:]+?)\s*\?\s*([^;\?:]+?)\s*:\s*([^;\?:]+?)\s*;/,
        cond, pre, new_cond, post,
        last_index = 0;
        
        // xxx again, in general we would need a full JS parser,
        // however the following should be fine in almost all practical cases.
        
        while (cond = rx.exec(ret.substr(last_index))) {

            log("xxx ret.substr(last_index)", ret.substr(last_index), ", cond:", cond);

            if (cond[1].indexOf("return") > -1) {
                last_index += cond.index + "return".length;
                continue;
            }            

            if (cond[1].indexOf(fname) > -1) {

                // Detect mistaken conditionals (not a "named let", and not a tail call)
                
                if (! (new RegExp('^\\s*\\(\\s*' + fname + '\\s*=\\s*')).test(cond[1])) {
                    log("expand_recursive_conditionals(): block: ", block);
                    log("expand_recursive_conditionals(): fname: ", fname);
                    log("expand_recursive_conditionals(): cond: ", cond);
                    throw new Error("expand_recursive_conditionals(block, fname) found a mistake: " +
                                    "to have a tail call, fname must *not* be in the test part of the conditional.");
                }

                // Do not modify irrelevant conditionals
                last_index += cond.index + cond[0].length;
                continue;
            }

            // Do not modify irrelevant conditionals
            
            if ((cond[2].indexOf(fname) < 0) && (cond[3].indexOf(fname)) < 0) {
                last_index += cond.index + cond[0].length;
                continue;
            }

            // Expand relevant conditionals
            
            // xxx permit without { } (this concerns more the "unroll_named_lets" code above)

            pre      = ret.substr(0, last_index + cond.index);
            new_cond = " if(" + cond[1] + "){return " + cond[2] + ";}return " + cond[3] + ";";
            post     = ret.substr(last_index + cond.index + cond[0].length);


            ret        = pre + new_cond + post;
            last_index = last_index + pre.length + new_cond.length;
        }

        return ret;
    };

})( /*global context object*/this );
